/* die orig. SW ist vom Hubi, wurde von mir(Ziyat T.) für den MI-WR abgeaendert.
Getestet auf ArduinoUno.
https://www.mikrocontroller.net/topic/525778
https://github.com/hm-soft/Hoymiles-DTU-Simulation
*/

#include <Arduino.h>
#include <SPI.h>
#include "CircularBuffer.h"
#include <RF24.h>
#include <RF24_config.h>
#include "hm_crc.h"
#include "hm_packets.h"
#include "Settings.h"     // Header für Einstellungen
//#include "DTSU.h"
#include "Debug.h"


#ifdef ESP8266
  #define DISABLE_EINT noInterrupts()
  #define ENABLE_EINT  interrupts()
#else     // für AVR z.B. ProMini oder Nano
  #define DISABLE_EINT EIMSK = 0x00
  #define ENABLE_EINT EIMSK = 0x01
#endif


#define RF_MAX_ADDR_WIDTH       (5) 
#define MAX_RF_PAYLOAD_SIZE     (32)

#ifdef ESP8266
#define PACKET_BUFFER_SIZE      (30) 
#else
#define PACKET_BUFFER_SIZE      (20) 
#endif

// Startup defaults until user reconfigures it
#define DEFAULT_RECV_CHANNEL    (3)             // 3 = Default channel for Hoymiles
//#define DEFAULT_SEND_CHANNEL  (75)            // 40 = Default channel for Hoymiles, 61
#define DEFAULT_RF_DATARATE     (RF24_250KBPS)  // Datarate

#include "NRF24_sniff_types.h"

static HM_Packets     hmPackets;
static uint32_t       tickMillis;


// Set up nRF24L01 radio on SPI bus plus CE/CS pins
// If more than one RF24 unit is used the another CS pin than 10 must be used
// This pin is used hard coded in SPI library
static RF24 radio1 (RF1_CE_PIN, RF1_CS_PIN);
static NRF24_packet_t bufferData[PACKET_BUFFER_SIZE];
static CircularBuffer<NRF24_packet_t> packetBuffer(bufferData, sizeof(bufferData) / sizeof(bufferData[0]));
static Serial_header_t SerialHdr;

#define CHECKCRC  1
static uint16_t lastCRC;
static uint16_t crc;

uint8_t         channels[]            = {3, 23, 40, 61, 75};   //{1, 3, 6, 9, 11, 23, 40, 61, 75}
uint8_t         channelIdx            = 1;                         // fange mit 23 an
uint8_t         DEFAULT_SEND_CHANNEL  = channels[channelIdx];      // = 23

static unsigned long timeLastPacket = millis();

// Function forward declaration
static void SendPacket(uint64_t dest, uint8_t *buf, uint8_t len);
//char * getChannelName (uint8_t i);
// static const int    ANZAHL_VALUES         = 16;
// static float        VALUES[ANZAHL_VALUES] = {};
// static const char   *CHANNEL_NAMES[ANZAHL_VALUES] 
//    = {"P1.Udc", "P1.Idc", "P1.Pdc", "P2.Udc", "P2.Idc", "P2.Pdc", 
//       "E-Woche", "E-Total", "E1-Tag", "E2-Tag", "Uac", "Freq.ac", "Pac", "E-heute", "Ipv", "WR-Temp"};
// static const uint8_t DIVISOR[ANZAHL_VALUES] = {10,100,10,10,100,10,1,1,1,1,10,100,10,0,0,10};
static const char BLANK = ' ';

// static boolean istTag = true;
// char CHANNELNAME_BUFFER[15];

#ifdef ESP8266
  #include "wifi.h"
  #include "ModWebserver.h"
  #include "Sonne.h"
#endif

//MI-WR Data
uint8_t sendBuf[MAX_RF_PAYLOAD_SIZE];
uint8_t MI_DataCMD = 0x36;   //begin with first PV 

//char WRdata[120];
float  U_DC =0;
float  I_DC =0;
float  U_AC =0;
float  F_AC =0;
float  P_DC =0;
float  Q_DC =0;           
float  TEMP =0;
uint8_t  STAT =0;
uint8_t  FCNT = 0;
uint8_t  FCODE = 0;
uint8_t  PV = 0;

//read from Serial for testing 
char SerialIn[10]="";
int Limit=0;
int SerCmd=0;
int OldLimit = Limit;
uint8_t WRInfo=1;



inline static void dumpData(uint8_t *p, int len) {
//------------------------------------------------------------------------------------------------
  while (len--){
    if (*p < 16)
      DEBUG_OUT.print(F("0"));
    DEBUG_OUT.print(*p++, HEX);
  }
  DEBUG_OUT.print(BLANK);
}//----dumpData------------------------------------------------------------------------------------

#ifdef ESP8266
IRAM_ATTR
#endif
void handleNrf1Irq() {
//---------------------------------------------------------------------------------------
  static uint8_t lostPacketCount = 0;
  uint8_t pipe;
  
  DISABLE_EINT;
  
  // Loop until RX buffer(s) contain no more packets.
  while (radio1.available(&pipe)) {
    if (!packetBuffer.full()) {
      NRF24_packet_t *p = packetBuffer.getFront();
      p->timestamp = micros(); // Micros does not increase in interrupt, but it can be used.
      p->packetsLost = lostPacketCount;
      uint8_t packetLen = radio1.getPayloadSize();
      if (packetLen > MAX_RF_PAYLOAD_SIZE)
        packetLen = MAX_RF_PAYLOAD_SIZE;

      radio1.read(p->packet, packetLen);
      packetBuffer.pushFront(p);
      lostPacketCount = 0;
      }
    else {
      // Buffer full. Increase lost packet counter.
      bool tx_ok, tx_fail, rx_ready;
      if (lostPacketCount < 255)
        lostPacketCount++;
      // Call 'whatHappened' to reset interrupt status.
      radio1.whatHappened(tx_ok, tx_fail, rx_ready);
      // Flush buffer to drop the packet.
      radio1.flush_rx();
      }
    }
  ENABLE_EINT;
}//---handleNrf1Irq-------------------------------------------------------------------

static void activateRFConf(void) {
//---------------------------------------------------------------------------------------
  radio1.setChannel(DEFAULT_RECV_CHANNEL);
  radio1.setDataRate(DEFAULT_RF_DATARATE);
  radio1.disableCRC();
  radio1.setAutoAck(0x00);
  radio1.setPayloadSize(MAX_RF_PAYLOAD_SIZE);
  radio1.setAddressWidth(5);
  radio1.openReadingPipe(1, DTU_RADIO_ID);

  // We want only RX irqs
  radio1.maskIRQ(true, true, false);

  // Use lo PA level, as a higher level will disturb CH340 DEBUG_OUT usb adapter
  radio1.setPALevel(RF24_PA_LOW);
  radio1.startListening();

  // Attach interrupt handler to NRF IRQ output. Overwrites any earlier handler.
  attachInterrupt(digitalPinToInterrupt(RF1_IRQ_PIN), handleNrf1Irq, FALLING); // NRF24 Irq pin is active low.

  // Initialize SerialHdr header's address member to promiscuous address.
  uint64_t addr = DTU_RADIO_ID;
  for (int8_t i = sizeof(SerialHdr.address) - 1; i >= 0; --i) {
    SerialHdr.address[i] = addr;
    addr >>= 8;
    } 

#ifndef ESP8266
  DEBUG_OUT.println(F("\nRadio Config:"));
  //radio1.printPrettyDetails();
  //radio1.printDetails();
  DEBUG_OUT.println();
#endif
  tickMillis = millis() + 200;
}
//--activateRFConf-------------------------------------------------------------------------------------

void setup(void) {
//---------------------------------------------------------------------------------------
  //Serial.begin(SER_BAUDRATE);
  DEBUG_OUT.begin(SER_BAUDRATE);
  DEBUG_OUT.flush();
  DEBUG_OUT.println(F("Setup: DTU for MI Simulation"));
  radio1.begin();

  // Disable shockburst for receiving and decode payload manually
  radio1.setAutoAck(false);
  radio1.setRetries(0, 0);

  // Configure nRF IRQ input
  pinMode(RF1_IRQ_PIN, INPUT);
  activateRFConf();

#ifdef ESP8266
  setupWifi();
  setupClock();
  setupWebServer();
//  setupUpdateByOTA();
  calcSunUpDown (getNow());
  istTag = isDayTime();
  DEBUG_OUT.print ("Es ist "); DEBUG_OUT.println (istTag?"Tag":"Nacht");
  hmPackets.SetUnixTimeStamp (getNow());
#else
  hmPackets.SetUnixTimeStamp(0x62456430);
#endif
// setupDTSU();

}//---setup------------------------------------------------------------------------------------

void isTime2Send () {
//-------------------------------------------------------------------------------------------
  int32_t size = 0;
  uint64_t dest = WR1_RADIO_ID;
  uint8_t UsrData[10]; 
  char Cmd = 0;  // Second timer

  if (millis() >= tickMillis) {
    static uint8_t telegram = 0;
    tickMillis += 200;    //200;
    //tickSec++; 
//    hmPackets.UnixTimeStampTick();
/*    if (++tickSec >= 5) {   // 5
      hmPackets.UnixTimeStampTick();
      tickSec = 0;
    } */

    if (telegram > 5) {   telegram = 0;    }       

    if (MI_DataCMD > 0x0039)
      MI_DataCMD= 0x0036;
    
    switch(telegram) {
      case 0:
        //set SubCmd and  UsrData
        if (Limit > 0){
          Cmd=0x51;          
          DEBUG_OUT.print(F("CMD 0x"));DEBUG_OUT.print(Cmd,HEX);DEBUG_OUT.print(F(" Sending Limit:")); DEBUG_OUT.println(Limit);
          UsrData[0]=0x5A;UsrData[1]=0x5A;UsrData[2]=100;//0x0a;// 10% limit   
          
          UsrData[3]=(Limit >> 8) & 0xFF;  DEBUG_OUT.print(F(" ")); UsrData[4]= Limit  & 0xFF; DEBUG_OUT.print(F(" "));   
          DEBUG_OUT.print(UsrData[3],DEC); DEBUG_OUT.print(UsrData[4],DEC);      
          size = hmPackets.GetCmdPacket((uint8_t *)&sendBuf, dest >> 8, DTU_RADIO_ID >> 8, Cmd, UsrData,5);
          //Limit=0;        
        }
        break;
      case 1: 
        if (SerCmd == 10001){
          Cmd=0x0f;
          UsrData[0]=0x01;
          DEBUG_OUT.print(F("CMD 0x"));DEBUG_OUT.print(Cmd,HEX);DEBUG_OUT.print(F(" Sending WRInfo Req:0x")); DEBUG_OUT.println(WRInfo,HEX);
          size = hmPackets.GetCmdPacket((uint8_t *)&sendBuf, dest >> 8, DTU_RADIO_ID >> 8, Cmd, UsrData,1);
        }      
        // code block
        break;
      default:
        UsrData[0]=0x0;//set SubCmd and  UsrData
        size = hmPackets.GetCmdPacket((uint8_t *)&sendBuf, dest >> 8, DTU_RADIO_ID >> 8, MI_DataCMD, UsrData,1);
    }
    SendPacket(dest, (uint8_t *)&sendBuf, size);
    telegram++;
    MI_DataCMD++;
/*    for (uint8_t warte = 0; warte < 2; warte++) {
      delay(1000);
      hmPackets.UnixTimeStampTick();
    }*/ 
  } //if millis
}//----isTime2Send---------------------------------------------------------------------------------------

void outputPacket(NRF24_packet_t *p, uint8_t payloadLen) {
//------------------------------------------------------------------------------------------------------
    // Write  packets lost, address and payload length
    dumpData((uint8_t *)&SerialHdr.packetsLost, sizeof(SerialHdr.packetsLost));
    dumpData((uint8_t *)&SerialHdr.address, sizeof(SerialHdr.address));
    // Trailing bit?!?
    dumpData(&p->packet[0], 2);
    // Payload length from PCF
    dumpData(&payloadLen, sizeof(payloadLen));
    // Packet control field - PID Packet identification
    uint8_t val = (p->packet[1] >> 1) & 0x03;
    DEBUG_OUT.print(val);
    DEBUG_OUT.print(F("  "));

    if (payloadLen > 9) {
      dumpData(&p->packet[2], 1);
      dumpData(&p->packet[3], 4);
      dumpData(&p->packet[7], 4);
      
      uint16_t remain = payloadLen - 2 - 1 - 4 - 4 + 4;

      if (remain < 32) {
        dumpData(&p->packet[11], remain);
        printf_P(PSTR("%04X "), crc);

        if (((crc >> 8) != p->packet[payloadLen + 2]) || ((crc & 0xFF) != p->packet[payloadLen + 3]))
          DEBUG_OUT.print(0);
        else
          DEBUG_OUT.print(1);
      }
      else {
        DEBUG_OUT.print(F("Ill remain "));
        DEBUG_OUT.print(remain);
      }
    }
    else {
      dumpData(&p->packet[2], payloadLen + 2);
      printf_P(PSTR("%04X "), crc);
    }
    DEBUG_OUT.println();
}//----outputPacket----------------------------------------------------------------------------

char InCnt=0;
void HandleSerialCmd(void){ //read from serial for WR control cmd's
//-------------------------------------------------------------------------------------------
  int temporary=0;
  
  if (Serial.available() > 0) {
    int incomingByte = Serial.read();
    if (incomingByte != 13){ //CR
      SerialIn[InCnt]=incomingByte; 
      InCnt++;       
    }     
    else {
      SerialIn[InCnt]=0;   //eofl
      temporary = atoi(SerialIn);
      if ((temporary >10000) || (temporary < 0))  //other cmds
        SerCmd = temporary;
      else{
        Limit = temporary*10; //Watt, with 1 dec
        OldLimit = Limit;
      }
//      #ifdef DEBUG_TX_DATA     HandleS
      DEBUG_OUT.println(SerialIn); 
      DEBUG_OUT.println(F(" Cmd:"));      
      DEBUG_OUT.println(temporary);
//      #endif 
      InCnt=0;
    }    
  } 
}//---HandleSerialCmd-------------------------------------------------------------------------------

void AnalyseMI1500(NRF24_packet_t *p){
//--------------------------------------------------------------------------------------------------
  static float TotalP[5]={0};

  switch (p->packet[2])  {
    case 0xD1:
      DEBUG_OUT.print (F("Limiting(0x51) is ok CMD=")); 
      DEBUG_OUT.println(p->packet[2], HEX);
      Limit=0; //stop sending Limit
      break;
    
    case 0x8f:
      DEBUG_OUT.print (F("WRInfo(0x8f) is ok CMD=")); 
      DEBUG_OUT.print(p->packet[2], HEX);DEBUG_OUT.print(F(" "));
      //WRSN
      DEBUG_OUT.print(p->packet[7], HEX);DEBUG_OUT.print(F(":"));
      DEBUG_OUT.print(p->packet[8], HEX);DEBUG_OUT.print(F(":"));
      DEBUG_OUT.print(p->packet[9], HEX);DEBUG_OUT.print(F(":"));
      DEBUG_OUT.print(p->packet[10], HEX);DEBUG_OUT.print(F(" "));
      //HW PN
      DEBUG_OUT.print(F("HWPN:"));
      DEBUG_OUT.print(p->packet[11], DEC);DEBUG_OUT.print(F("-"));
      DEBUG_OUT.print(p->packet[12], DEC);DEBUG_OUT.print(F("-"));
      DEBUG_OUT.print(p->packet[13], DEC);DEBUG_OUT.print(F(" "));
      //HW Vers      
      DEBUG_OUT.print(F(" HWVers:"));
      DEBUG_OUT.print(p->packet[14], DEC);DEBUG_OUT.print(F("."));
      DEBUG_OUT.print(p->packet[15], DEC);DEBUG_OUT.print(F(" "));
      //APPFWBuild_VER
      DEBUG_OUT.print(F(" APPFVers:"));
      DEBUG_OUT.print(p->packet[16], DEC);DEBUG_OUT.print(F("."));
      DEBUG_OUT.print(p->packet[17], DEC);DEBUG_OUT.print(F("."));
      DEBUG_OUT.print(p->packet[18], DEC);DEBUG_OUT.print(F(" "));
      //GPFCode
      DEBUG_OUT.print(F(" GPFCode:"));DEBUG_OUT.print(((p->packet[19]<<8) + p->packet[20]), DEC);
      //GPFVer      
      DEBUG_OUT.print(F(" GPFVers:"));DEBUG_OUT.print(((p->packet[21]<<8) + p->packet[22]), DEC);

      SerCmd=0; //stop sending WRInfo
      break;    
    default:
      U_DC =  (float) ((p->packet[11] << 8) + p->packet[12])/10;
      I_DC =  (float) ((p->packet[13] << 8) + p->packet[14])/10;
      U_AC =  (float) ((p->packet[15] << 8) + p->packet[16])/10;
      F_AC =  (float) ((p->packet[17] << 8) + p->packet[18])/100;
      P_DC =  (float)((p->packet[19] << 8) + p->packet[20])/10;
      Q_DC =  (float)((p->packet[21] << 8) + p->packet[22])/1;           
      TEMP =  (float) ((p->packet[23] << 8) + p->packet[24])/10;

      STAT = (uint8_t)(p->packet[25] );
      FCNT = (uint8_t)(p->packet[26]);
      FCODE = (uint8_t)(p->packet[27]);

      if (p->packet[2] == 0xB6)  {PV= 1; TotalP[1]=P_DC;}
      if (p->packet[2] == 0xB7)  {PV= 2; TotalP[2]=P_DC;}//port 2
      if (p->packet[2] == 0xB8)  {PV= 3; TotalP[3]=P_DC;}//port 3
      if (p->packet[2] == 0xB9)  {PV= 4; TotalP[4]=P_DC;}//port 4
      TotalP[0]=TotalP[1]+TotalP[2]+TotalP[3]+TotalP[4];
      
  //    arduino problem solving sprintf float
      // dtostrf(volt, 3, 1, str_temp); //3 long 1 decpoint
      // sprintf(LcdStr,"%5sV", str_temp);
        
      // sprintf (WRdata,"PV%1d %sV %.1f A %.1f V %.1f Hz %.1f W %.1f Wh %3.1f C Sts:%1d Cnt:%1d ",
      //          PV,String(U_DC),I_DC,U_AC,F_AC,P_DC,Q_DC,TEMP,STAT,CNT);
      // DEBUG_OUT.print(WRdata);
      DEBUG_OUT.print(F(""));DEBUG_OUT.print(TotalP[0],1); DEBUG_OUT.print(F("W "));
      DEBUG_OUT.print(F("PV:"));DEBUG_OUT.print(PV,DEC); DEBUG_OUT.print(F(" "));
      DEBUG_OUT.print(P_DC,1); DEBUG_OUT.print(F("W "));
      DEBUG_OUT.print(U_DC,1); DEBUG_OUT.print(F("V "));
      DEBUG_OUT.print(I_DC,1); DEBUG_OUT.print(F("A "));
      DEBUG_OUT.print(Q_DC,1); DEBUG_OUT.print(F("Wh "));
      DEBUG_OUT.print(U_AC,1); DEBUG_OUT.print(F("V "));
      DEBUG_OUT.print(F_AC,1); DEBUG_OUT.print(F("Hz "));
      DEBUG_OUT.print(TEMP,1); DEBUG_OUT.print(F("C "));
      DEBUG_OUT.print(F(" Sts:")); DEBUG_OUT.print(STAT,DEC); 
      DEBUG_OUT.print(F(" FCnt:"));DEBUG_OUT.print(FCNT,DEC); 
      DEBUG_OUT.print(F(" FCd:"));DEBUG_OUT.print(FCODE,DEC); 
      DEBUG_OUT.print(F(" OldL:"));DEBUG_OUT.print(OldLimit,DEC);     

  }
}//analayseMI1500----------------------------------------------------------------------------------

void ReadRFandAnalyse(void) {
//--------------------------------------------------------------------------------------------------
  while (!packetBuffer.empty()) {
    timeLastPacket = millis();
    // One or more records present
    NRF24_packet_t *p = packetBuffer.getBack();

    // Shift payload data due to 9-bit packet control field
    for (int16_t j = sizeof(p->packet) - 1; j >= 0; j--) {
     if (j > 0)
        p->packet[j] = (byte)(p->packet[j] >> 7) | (byte)(p->packet[j - 1] << 1);
     else
        p->packet[j] = (byte)(p->packet[j] >> 7);
     }

    SerialHdr.timestamp   = p->timestamp;
    SerialHdr.packetsLost = p->packetsLost;
    // Check CRC
    crc = 0xFFFF;
    crc = crc16((uint8_t *)&SerialHdr.address, sizeof(SerialHdr.address), crc, 0, BYTES_TO_BITS(sizeof(SerialHdr.address)));
    // Payload length
    uint8_t payloadLen = ((p->packet[0] & 0x01) << 5) | (p->packet[1] >> 3);
    // Add one byte and one bit for 9-bit packet control field
    crc = crc16((uint8_t *)&p->packet[0], sizeof(p->packet), crc, 7, BYTES_TO_BITS(payloadLen + 1) + 1);

    if (CHECKCRC) {
      // If CRC is invalid only show lost packets
      if (((crc >> 8) != p->packet[payloadLen + 2]) || ((crc & 0xFF) != p->packet[payloadLen + 3])) {
        if (p->packetsLost > 0) {
          DEBUG_OUT.print(F(" Lost: "));
          DEBUG_OUT.println(p->packetsLost);
        }
        packetBuffer.popBack();
        continue;
      }
      // Dump a decoded packet only once
      if (lastCRC == crc) {
        packetBuffer.popBack();
        continue;
      }
      lastCRC = crc;
    }// if checkcrc
    // Don't dump mysterious ack packages
    if (payloadLen == 0) {
      packetBuffer.popBack();
      continue;
    }

    if (DEBUG)
      outputPacket (p, payloadLen); //output received data
    
    //toe analyse
    AnalyseMI1500(p);

    if (p->packetsLost > 0) {
      DEBUG_OUT.print(F(" Lost: "));
      DEBUG_OUT.println(p->packetsLost);
    }
    DEBUG_OUT.println();
    // Remove record as we're done with it.
    packetBuffer.popBack();
  } //while (!packetBuffer.empty())  GET PACKET
 
}
//-----ReadRFandAnalyse-------------------------------------------------------------------------------

static void SendPacket(uint64_t dest, uint8_t *buf, uint8_t len) {
//----------------------------------------------------------------------------------------------------
  DISABLE_EINT;
  radio1.stopListening();

#ifdef CHANNEL_HOP
  static uint8_t hop = 0;
  #if DEBUG_TX_DATA   
    DEBUG_OUT.print(F("Send... CH"));
    DEBUG_OUT.print(channels[hop]);
    DEBUG_OUT.print(F(" "));
  #endif  
  radio1.setChannel(channels[hop++]);
  if (hop >= sizeof(channels) / sizeof(channels[0]))
    hop = 0;
#else
  radio1.setChannel(DEFAULT_SEND_CHANNEL);
#endif

#if DEBUG_TX_DATA
//toe  send packet output
  for (uint8_t i = 0; i < len; i++){
      if (buf[i]==0){DEBUG_OUT.print(F("00"));}
      else {DEBUG_OUT.print(buf[i],HEX);} 
        }
  DEBUG_OUT.println (F(""));
//toe
#endif

  radio1.openWritingPipe(dest);
  radio1.setCRCLength(RF24_CRC_16);
  radio1.enableDynamicPayloads();
  radio1.setAutoAck(true);
  radio1.setRetries(3, 15);

  radio1.write(buf, len);

  // Try to avoid zero payload acks (has no effect)
  radio1.openWritingPipe(DUMMY_RADIO_ID);

  radio1.setAutoAck(false);
  radio1.setRetries(0, 0);
  radio1.disableDynamicPayloads();
  radio1.setCRCLength(RF24_CRC_DISABLED);

  radio1.setChannel(DEFAULT_RECV_CHANNEL);
  radio1.startListening();
  ENABLE_EINT;
}
//----SendPacket---------------------------------------------------------------------------------

void loop(void) {
//===============================================================================================
  float P_DTSU=0;
  ReadRFandAnalyse();
  HandleSerialCmd();  //read from serial any command
  //toe if (istTag) 
  isTime2Send();//SEND PACKET
  //--------------------------------------------------  
  //P_DTSU=loopDTSU(); //read DTSU666 Power

  #ifdef ESP8266
    checkWifi();
    webserverHandle();
    //checkUpdateByOTA();
    if (hour() == 0 && minute() == 0) {
      calcSunUpDown(getNow());  
    }
    if (minute() % 15 == 0 && second () == 0) {  // alle 15 Minuten neu berechnen ob noch hell
      istTag = isDayTime();
      DEBUG_OUT.print (F("Es ist ")); DEBUG_OUT.println (F(istTag?"Tag":"Nacht"));
    }
  #endif

}//-----loop-----------------------------------------------------------------------------------------
